
#include "proto.h"

int i2c_attach_slave_read_interrupt(i2c_dev_t *dev)
{
	if (-1 == ThreadCtl(_NTO_TCTL_IO, 0))
	{
		if ( uiVerbosity )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR, "%s:i2c_attach_slave_interrupt:ThreadCtl: %s.", pszProgram, strerror(errno) );
		return NULL;
	}

	/* Initialize interrupt handler */
	SIGEV_INTR_INIT(&dev->slave_read_intrevent);

	/*InterruptAttachEvent() tells the kernel that the event (intrevent) should be
	 returned whenever the interrupt (intr) is detected, and that the interrupt level should be
	 masked off. */
	dev->slave_read_iid = InterruptAttachEvent(dev->intr, &dev->slave_read_intrevent, _NTO_INTR_FLAGS_TRK_MSK);
	if (dev->slave_read_iid == -1)
	{
		if ( uiVerbosity )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR, "%s:i2c_attach_slave_read_interrupt:InterruptAttachEvent: %s.", pszProgram, strerror(errno) );
		return -1;
	}
	/* enable interrupts */
//	out16(dev->regbase + I2C_CTRREG_OFF, in16(dev->regbase + I2C_CTRREG_OFF) | CTRREG_IEN | CTRREG_IIEN );
	if (uiVerbosity >= LOGGING_PROGRESS)
		slogf(_SLOGC_SNELL_IQMIT_I2C_RESMGR, _SLOG_INFO, "%s:i2c_attach_slave_read_interrupt: Successful", pszProgram);
	return 0;
}

int i2c_deattach_slave_read_interrupt(i2c_dev_t *dev)
{
	int r;

	if (-1 == ThreadCtl(_NTO_TCTL_IO, 0))
	{
		if ( uiVerbosity )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR, "%s:i2c_deattach_slave_read_interrupt:ThreadCtl: %s.", pszProgram, strerror(errno) );
		return NULL;
	}
	r = InterruptDetach(dev->slave_read_iid);
	if (r == -1)
	{
		if ( uiVerbosity )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_WARNING, "%s:i2c_deattach_slave_read_interrupt:InterruptDetach: %s.",
					pszProgram, strerror(errno) );
	}
	else if (uiVerbosity >= LOGGING_PROGRESS)
		slogf(_SLOGC_SNELL_IQMIT_I2C_RESMGR, _SLOG_INFO, "%s:i2c_deattach_slave_read_interrupt: Successful", pszProgram);
	return 0;
}

int i2c_attach_slave_write_interrupt(i2c_dev_t *dev)
{
	if (-1 == ThreadCtl(_NTO_TCTL_IO, 0))
	{
		if ( uiVerbosity )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR, "%s:i2c_attach_slave_write_interrupt:ThreadCtl: %s.",
					pszProgram, strerror(errno) );
		return NULL;
	}

	// Initialize interrupt handler
	SIGEV_INTR_INIT(&dev->slave_write_intrevent);

	// InterruptAttachEvent() tells the kernel that the event (intrevent) should be
	// returned whenever the interrupt (intr) is detected, and that the interrupt level should be
	// masked off.
	dev->slave_write_iid = InterruptAttachEvent(dev->intr, &dev->slave_write_intrevent, _NTO_INTR_FLAGS_TRK_MSK);
	if (dev->slave_write_iid == -1)
	{
		if ( uiVerbosity )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR, "%s:i2c_attach_slave_write_interrupt:InterruptAttachEvent: %s.",
					pszProgram, strerror(errno) );
		return -1;
	}

	// enable interrupts
//	out16(dev->regbase + I2C_CTRREG_OFF, in16(dev->regbase + I2C_CTRREG_OFF) | CTRREG_IEN | CTRREG_IIEN );
	if (uiVerbosity >= LOGGING_PROGRESS)
		slogf(_SLOGC_SNELL_IQMIT_I2C_RESMGR, _SLOG_INFO, "%s:i2c_attach_slave_write_interrupt: Successful", pszProgram);
	return 0;
}

int i2c_deattach_slave_write_interrupt(i2c_dev_t *dev)
{
	int r;

	if (-1 == ThreadCtl(_NTO_TCTL_IO, 0))
	{
		if ( uiVerbosity )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR, "%s:i2c_deattach_slave_write_interruptt:ThreadCtl: %s.",
					pszProgram, strerror(errno) );
		return NULL;
	}
	r = InterruptDetach(dev->slave_write_iid);
	if (r == -1)
	{
		if ( uiVerbosity )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_WARNING, "%s:i2c_deattach_slave_write_interrupt:InterruptDetach: %s.",
					pszProgram, strerror(errno) );
	}
	else if (uiVerbosity >= LOGGING_PROGRESS)
		slogf(_SLOGC_SNELL_IQMIT_I2C_RESMGR, _SLOG_INFO, "%s:i2c_deattach_slave_write_interrupt: Successful", pszProgram);
	return 0;
}

/* We could add locks if needed */
char szlibHex[] = "0123456789ABCDEF";

void fnlibConvertFromHex( char* pszText, char* pszHex, int iLength )
{
	char *pszPtr1;
	char *pszPtr2;

	pszPtr1 = pszText;
	pszPtr2 = pszHex;
//	*pszPtr1++ = '0';
//	*pszPtr1++ = 'x';
	while ( iLength-- )
	{
		int i1 = *pszPtr2++;
		*pszPtr1++ = szlibHex[ ( i1 & 0xF0 ) >> 4 ];
		*pszPtr1++ = szlibHex[ i1 & 0x0F ];
	}
	*pszPtr1 = 0;
}

int fnlibDisplayData( char psz3Lines[][80], int iLinesSize, char** pcData, int* piLen )
{
	int i1, i2;
	char szHex[256];
	i2 = *piLen;
	if ( iLinesSize < 80*3 )
		return 0;
	if ( ( i1 = *piLen ) )
	{
		if ( i1 > 64 )
			i1 = 64;
		fnlibConvertFromHex( szHex, *pcData, i1 );
		char *pszHex = szHex;
		char *pszLine1 = psz3Lines[0];
		char *pszLine2 = psz3Lines[1];
		char *pszLine3 = psz3Lines[2];
		for ( i2 = 0; i2 < i1; )
		{
			*pszLine1++ = ( isprint( **pcData ) ? **pcData : ' ' );
			(*pcData)++;
			*pszLine2++ = *pszHex++;
			*pszLine3++ = *pszHex++;
			if ( i2++ && (i2 < i1) && ( (i2 % 4 ) == 0 ) )
				*pszLine1++ = *pszLine2++ = *pszLine3++ = ' ';
		}
		*pszLine1++ = *pszLine2++ = *pszLine3++ = '\0';
		*piLen -= i1;
	}
	return ( i1 );
}

i2c_status_t i2c_set_callback(i2c_dev_t *dev, void *buf, unsigned int len)
{
	memcpy(dev->slave.data, buf, len);
	dev->slave.len = len;
	if (uiVerbosity >= LOGGING_PROGRESS)
	{
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_INFO,
				"%s:i2c_set_callback: Slave transmit length %d bytes:", pszProgram, len );
		char sz3Lines[3][80];
		int i = len;
		char *pch = buf;
		while( fnlibDisplayData( &sz3Lines[0], sizeof(sz3Lines), &pch, &i ) )
		{
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_INFO, "%s:i2c_set_callback: %s", pszProgram, sz3Lines[0] );
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_INFO, "%s:i2c_set_callback: %s", pszProgram, sz3Lines[1] );
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_INFO, "%s:i2c_set_callback: %s", pszProgram, sz3Lines[2] );
		}
	}
	return 0;
}

i2c_status_t i2c_slave_read(i2c_dev_t *dev, void *buf, unsigned int len,
		unsigned int stop)
{
	i2c_status_t status; //needs to come from an interrupt
	i2c_status_t ret;
	int t;

	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
				"%s:i2c_slave_read: entry", pszProgram);
	if (len <= 0)
	{
		if (uiVerbosity >= LOGGING_DEBUG_FINE)
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
					"%s:i2c_slave_read: length passed as %d so return 0",
					pszProgram, len);
		return 0;
	}
	t = i2c_attach_slave_read_interrupt(dev);
	if (t == -1)
	{
		if (uiVerbosity >= LOGGING_ERRORS)
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR,
					"%s:i2c_slave_read:i2c_attach_slave_read_interrupt failure",
					pszProgram);
		return (t);
	}
	/* Wait for an interrupt */
//start:
	status = i2c_slave_wait_forever(dev, AM335x_I2C_RRDY);	// Signal a slave read wait forever

	/*  If we are Addressed as a slave we keep control until we have left the i2c_slave function,
	 *  lock is then released outside of this.
	 *
	 *  If we are addresses as a slave after we have started a Master Write or Master Read
	 *  and have not fully completed the  Master Write  / Master Read action (this means that the Master has not released
	 *  the control lock and this could block here) . Hence we "try" the lock and if we can't lock it we just ignore this slave request
	 *  and go back to waiting for another one, while the Master Write finishes and releases the lock.
	 *
	 *  This should be very unlikely to happen since the only reason the modules does a Master Write is in response to a previous
	 *  Master Read by the GW . The GW would need to do a Master Write to the Module, and then do a Master Read (Polling), while the Module
	 *  was attempting to complete it's Master Write  */
	/*
	t = pthread_mutex_trylock(&dev->control_lock);
	if (t == EBUSY)
	{
		if ( uiVerbosity >= LOGGING_DEBUG_FINE )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
				"%s:i2c_slave_read:pthread_mutex_trylock: mutex already locked, ignore this slave request wait for another so retry", pszProgram);
		goto start;
	}
	else if ( uiVerbosity >= LOGGING_DEBUG_FINE )
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
			"%s:i2c_slave_read:pthread_mutex_trylock: status 0x%X", pszProgram, t);
	 */

	t = i2c_deattach_slave_read_interrupt((void *) dev);
	if (t == -1)
	{
		if (uiVerbosity >= LOGGING_ERRORS)
			slogf(_SLOGC_SNELL_IQMIT_I2C_RESMGR, _SLOG_ERROR,
					"%s:i2c_slave_read:i2c_deattach_slave_read_interrupt failure", pszProgram);
		return(t);
	}
	if ( status )
	{
		if (uiVerbosity >= LOGGING_DEBUG_FINE)
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
					"%s:i2c_slave_read:i2c_slave_wait_forever: error status 0x%X return -1", pszProgram, status);

		return (-1);
	}
	// Slave Rx/ Read mode as Master Writing -eg Master Tx
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
				"%s:i2c_slave_read:  Starting receive (CTRREG 0x%X STSREG 0x%X)", pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
/*
	//A dummy read of I2C_I2DR in Slave Receive mode releases SCL, allowing the master to send data
	int c = in16(dev->regbase + I2C_DATREG_OFF) & 0x0FF;
	printf( "%X %c\n", c, c );
*/
	// Move to waiting for the data cycle now
	ret = i2c_slave_recv(dev, buf, len, stop);
//	out16(dev->regbase + AM335x_I2C_SBLOCK, 0 );					// Tell it to only use OA0
	out16( dev->regbase + AM335x_I2C_STAT, in16( dev->regbase + I2C_STSREG_OFF ) & ( AM335x_I2C_RRDY | AM335x_I2C_ROVR | AM335x_I2C_ARDY | AM335x_I2C_NACK  | AM335x_I2C_AL | AM335x_I2C_BF ) );
	status = errno;
	//return from the Slave read
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
				"%s:i2c_slave:i2c_slave_recv: return 0x%X %s", pszProgram,
				ret, ret > 0 ? "" : strerror(t));
	// Unlock the control now that the slave function has returned
	/*
	t = pthread_mutex_unlock(&dev->control_lock);
	if (t != 0)
	{
		if (uiVerbosity >= LOGGING_ERRORS)
			slogf(_SLOGC_SNELL_IQMIT_I2C_RESMGR, _SLOG_ERROR,
				"%s:io_read:pthread_mutex_unlock failure", pszProgram);
		return(-1);
	}
	 */
	errno = status;
//	i2c_reset(dev);
	return ret;
}

void i2c_slave_write(i2c_dev_t *dev)
{
	i2c_status_t status; //needs to come from an interrupt
	int t;

	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
				"%s:i2c_slave_write: entry", pszProgram);
	t = i2c_attach_slave_write_interrupt(dev);
	if (t == -1)
	{
		if (uiVerbosity >= LOGGING_ERRORS)
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR,
					"%s:i2c_slave_write:i2c_attach_slave_write_interrupt failure",
					pszProgram);
		exit (t);
	}
	/* Wait for an interrupt */
start:
	status = i2c_slave_wait_forever(dev, AM335x_I2C_XRDY);	// Signal a slave write wait forever

	/*  If we are Addressed as a slave we keep control until we have left the i2c_slave function,
	 *  lock is then released outside of this.
	 *
	 *  If we are addresses as a slave after we have started a Master Write or Master Read
	 *  and have not fully completed the  Master Write  / Master Read action (this means that the Master has not released
	 *  the control lock and this could block here) . Hence we "try" the lock and if we can't lock it we just ignore this slave request
	 *  and go back to waiting for another one, while the Master Write finishes and releases the lock.
	 *
	 *  This should be very unlikely to happen since the only reason the modules does a Master Write is in response to a previous
	 *  Master Read by the GW . The GW would need to do a Master Write to the Module, and then do a Master Read (Polling), while the Module
	 *  was attempting to complete it's Master Write  */
/*
	t = pthread_mutex_trylock(&dev->control_lock);
	if (t == EBUSY)
	{
		if ( uiVerbosity >= LOGGING_DEBUG_FINE )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
					"%s:i2c_slave_write:pthread_mutex_trylock: mutex already locked, ignore this slave request wait for another so retry", pszProgram);
		goto start;
	}
	else if ( uiVerbosity >= LOGGING_DEBUG_FINE )
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
				"%s:i2c_slave_write:pthread_mutex_trylock: status 0x%X", pszProgram, t);
*/
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
				"%s:i2c_slave_write:i2c_slave_wait_forever: status 0x%X", pszProgram, status);
	if (status)
		goto start;
	// Slave Tx/ Write mode as Master Reading eg Master Rx
	if ( uiVerbosity >= LOGGING_DEBUG_FINE )
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
				"%s:i2c_slave_write: Slave Write", pszProgram);
/*sws
	// Make sure I2C Module on and Interrupts CTRREG_IIEN on
	out16(dev->regbase + I2C_CTRREG_OFF, CTRREG_IEN | CTRREG_IIEN);
	// Slave mode
	out16(dev->regbase + I2C_CTRREG_OFF, in16(dev->regbase + I2C_CTRREG_OFF) & (~CTRREG_MSTA));
	//Set Tx Mode based on SRW when STSREG_IAAS is set only (eg when addresses)  - here we are a  Transmitter CTRREG_MTX
	out16(dev->regbase + I2C_CTRREG_OFF, in16(dev->regbase + I2C_CTRREG_OFF) | CTRREG_MTX);
	// Send Acknowledgement ~CTRREG_TXAK
	out16(dev->regbase + I2C_CTRREG_OFF, in16(dev->regbase + I2C_CTRREG_OFF) & (~CTRREG_TXAK));
	//Repeat Start =0 N/A here
	out16(dev->regbase + I2C_CTRREG_OFF, in16(dev->regbase + I2C_CTRREG_OFF) & (~CTRREG_RSTA));
*/

	// Slave Tx bytes
	i2c_slave_send(dev);
	status = in16( dev->regbase + I2C_STSREG_OFF );
//	printf( "Slave Write done status 0x%X\n", status );
	out16( dev->regbase + AM335x_I2C_STAT,  status & ( AM335x_I2C_XRDY | AM335x_I2C_XUDF | AM335x_I2C_ARDY | AM335x_I2C_NACK  | AM335x_I2C_AL | AM335x_I2C_BF ) );

	/* After the Slave Write, we don't want to return from io_read , we only do this after a Slave Read
	 * hence we return to waiting for another interrupt.
	 * Additionally we have to unlock control here instead of in the io_read function, now that the slave has finished
	 * and we don't leave i2c_slave
	 */
/*
	t = pthread_mutex_unlock(&dev->control_lock);
	if (t != 0)
	{
		if ( uiVerbosity >= LOGGING_ERRORS )
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR,
					"%s:i2c_slave_write:pthread_mutex_unlock: failure exit 1", pszProgram);
		exit(1);
	}
*/
	if ( uiVerbosity >= LOGGING_DEBUG_FINE )
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR,
				"%s:i2c_slave_write: complete, so do it again Status In 0x%X Out 0x%X", pszProgram, status, in16( dev->regbase + I2C_STSREG_OFF ) );
//	i2c_reset(dev);		// reset and restart
	goto start;
	t = i2c_deattach_slave_write_interrupt((void *) dev);
	if (t == -1)
	{
		if (uiVerbosity >= LOGGING_ERRORS)
			slogf(_SLOGC_SNELL_IQMIT_I2C_RESMGR, _SLOG_ERROR,
					"%s:i2c_slave_write:i2c_deattach_slave_write_interrupt failure", pszProgram);
		return;
	}
	/* Should never get here */
	if ( uiVerbosity >= LOGGING_ERRORS )
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_CRITICAL,
				"%s:i2c_slave_write: Should never get here return 0", pszProgram);
	return;
}
